//
// Harris Corner Detector
//

#include	<math.h>
#include	"Image.h"
#include	"Harris.h"
#include	"HarrisTools.h"

#define		MaxGray		255

// Harris corner detector w/o output image
int
Harris(const Image &InImg,
       int **up0, int **vp0, double ** value, int *cnr_num,
       int Ns, double sigma, double rk,
       double s, int max_corner, double min_dist)
{
  double	r, d;
  int		Ix, Iy;
  double	Ix2, Iy2, Ixy;
  int		gx, gy;
  double	gxx, gyy, gxy;
  double	max_r, r_thd;
  int		corner_num;
  int		x, x0, xx, u0, v0;
  int		offwidth, offheight;
  int		uoff, voff;//оݤȤʤħΰֺɸ
  int		dblw;
  
  int		*up = new int [max_corner];
  int		*vp = new int [max_corner];

  int		width = InImg.xl();
  int		height = InImg.yl();
//   int		*Im   = InImg.data;
   int		*Im   = new int [width * height];

  double	*Gw   = new double [(2*Ns+1) * (2*Ns+1)];
  int		*Imx  = new int [width * height];
  int		*Imy  = new int [width * height];
  int		*Imxx = new int [width * height];
  int		*Imyy = new int [width * height];
  int		*Imxy = new int [width * height];
  float 	*out  = new float [width * height];
  double	*dp   = new double [max_corner];

   // memory allocation check
  if (!Im || !up || !vp || !Gw || !Imx || !Imy || !Imxx || !Imyy || !Imxy ||
	  !out || !dp)
	return 0;

  
  for(int i = 0; i<width;i++){
	for(int j=0; j<height; j++){
	  Im[j*width + i] = InImg.rget(i,j);
	}
  }
  (*up0) = up;
  (*vp0) = vp;
   
  make_Gauss_window(Gw, Ns, sigma);

  dblw = width * 2;//12Ȥưᡣ2ȤƤϡvoffset Υɽ

  voff = uoff = 2;
  offwidth = width - uoff;
  offheight = height - voff;
   
  
  //γƲǤˤu,v줾ʬͤʬˤ롣
  //ƲǤˤơ+-2βͤκ + +-1βǤκ
  //ƲǤˤơ+-2βͤκ + +-1βǤκ
  //ϲ岼2ǤľʬФƷ׻
  for (v0 = voff; v0 < offheight; v0++)	{
	  x = (v0 * width) + uoff;//ʬκɸͷ׻
	  for (u0 = uoff; u0 < offwidth; u0++, x++)
	  {
		Ix = 2 * (Im[x + 2] - Im[x - 2]) + (Im[x + 1] - Im[x - 1]);
		Iy = 2 * (Im[x + dblw] - Im[x - dblw])
			  + (Im[x + width] - Im[x - width]);
	  
		Imx[x] = Ix;
		Imy[x] = Iy;
	  }
  }

  //ʬ᤿ʬꤵ˾岼2ǤŤľʬФơ
  uoff += 2;
  voff += 2;
  offwidth -= 2;
  offheight -= 2;
  
  
  //(оݤȤƤǤ濴Ȥ5*5ΰˤ)1ʬͤ2׻
  for (v0 = voff; v0 < offheight; v0++)
  {
	x = (v0 * width) + uoff;
	for (u0 = uoff; u0 < offwidth; u0++, x++)
	{
	  gx = Imx[x - 2] + Imx[x - 1] + Imx[x] + Imx[x + 1] + Imx[x + 2];
	  gy = Imy[x - dblw] + Imy[x - width] + Imy[x]
		  + Imy[x + width] + Imy[x + dblw];
	  Imxx[x] = gx * gx;
	  Imyy[x] = gy * gy;
	  Imxy[x] = gx * gy;
	}
  }

  //ʿ경Ԥΰ礭ͤơ׻оΰ򶹤롣
  uoff += Ns;
  voff += Ns;
  offwidth -= Ns;
  offheight -= Ns;
  max_r = 0.0;
  
  //ʬͤ˥ʿ경Ԥ(2ʬѤȤ)
  for (v0 = voff; v0 < offheight; v0++)
  {
	x = v0 * width + uoff;
	for (u0 = uoff; u0 < offwidth; u0++, x++)
	{
	  gxx = gyy = gxy = 0.0;
	  xx = 0;
	  for (int vv = -Ns; vv <= Ns; vv++)
	  {
		x0 = (v0 + vv) * width + u0;
		for (int uu = -Ns; uu <= Ns; uu++)
		{
		  gxx += Gw[xx] * Imxx[x0 + uu];
		  gyy += Gw[xx] * Imyy[x0 + uu];
		  gxy += Gw[xx] * Imxy[x0 + uu];
		  xx++;
		}
	  }
  
	  Ix2 = gxx * gxx;
	  Iy2 = gyy * gyy;
	  Ixy = gxy * gxy;
	  r = (Ix2 * Iy2 - sqr(Ixy)) - rk * sqr(Ix2 + Iy2);

	  if (r > max_r)
		max_r = r;
  
	  out[x] = r;
	}
  }

  r_thd = max_r * s;

  uoff += 2;
  voff += 2;
  offwidth -= 2;
  offheight -= 2;

  corner_num = 0;
  for (v0 = voff; v0 < offheight; v0++)
  {
	  x = v0 * width + uoff;
	  for (u0 = uoff; u0 < offwidth; u0++, x++)
	  {
		d = out[x];
		if (d <= r_thd)  continue;

		if (   check_ge_upperleft(d, out, x, width)
			  && check_gt_lowerright(d, out, x, width))
		{
		  corner_num = add_list(u0, v0, d, up, vp, dp,
					  corner_num, max_corner, min_dist);
		}
	  }
  }
  
  //ϥꥹڥ졼ͤݻ
  (*value) = new double [corner_num];
  if(!(*value ))
	return 0;
  for(int i=0; i<corner_num; i++)
	(*value)[i] = dp[i];
	

// delete arrays
  delete [] dp;
  delete [] out;
  delete [] Imxy;
  delete [] Imyy;
  delete [] Imxx;
  delete [] Imy;
  delete [] Imx;
  delete [] Gw;
  
  return (*cnr_num = corner_num);
}
