/************************************************
 Image.cc   Ver 1.0

             created  by T.Sakamoto (2000/09/11) 
         modified by H.Kamo     (2000/10/02)
         modified by H.Kamo     (2000/10/06)
         modified by T.Sakamoto (2001/05/27)
         modified by Y.Nakano   (2005/04/06)
         modified by K.Suzukawa (2005/11/22)
*************************************************/

#include <math.h>
#if __GNUC__ < 3
# include <iostream.h>
# include <stream.h>
#else
# include <iostream>
# include <iomanip>
# include <new>
using std::cerr;
using std::endl;
using std::hex;
using std::setw;
#endif
#include "Image.h"
#include "ScreenPS.h"
// #include "PStools.h"


//  int head_jump(FILE *fp, char *magic, int &xs, int &ys, int &maxlevel)
//      :  եݥ, ޥåʥС (output),
//             ν,  (output), 統 (output)
//      :   1,  0
//      :  եΥإåʬɤ߹ߴؿ
static int head_jump(FILE *fp, char *magic, int &xs, int &ys, int &maxlevel)
{
  int   count = 0;
  char  c;
  int   flg1 = 0, flg2 = 1;
  char  buff[BUFSIZ];
  int   i = 0;

  while (count < 4){
    c = fgetc(fp);
    if (c == '#' || c == ' ' || c == '\n' || c == '\t') {
      flg2 = 0;
      if (c == '#')
    while (c != '\n') c = fgetc(fp);
    }
    else flg2 = 1;
    
    if (flg2 == 1) {
      if (!isalnum(c))
    return 0;
      buff[i] = c;
      i++;
    }
    if (flg1 == 1 && flg2 == 0) {
      buff[i] = ' ';
      count++;
      i++;
    }
    flg1 = flg2;
  }
  buff[i] = '\0';
  sscanf(buff, "%s %d %d %d", magic, &xs, &ys, &maxlevel);
  
  return 1;
}



//////////////////
//
//  Image 饹
//
//////////////////

//  󥹥ȥ饯  Image(int xs, int ys)
//      :  ν, 
Image::Image(int xs, int ys)
{
  assert( xs > 0 && ys > 0);

  width   = xs;
  height  = ys;
  xcenter = xs / 2;
  ycenter = ys / 2;

#if __GNUC__ < 3
  data = new int [ys * xs];
#else
  data = new(std::nothrow) int [ys * xs];
#endif
  
  if (data == 0) {
    cerr << "cannot allocate memories (Image)." << endl;
    exit(1);
  }
}


//  󥹥ȥ饯  Image(const char *filename)
//      :  pgmΥե̾
Image::Image(const char *filename)
{
  width = height = -1;
  data = NULL;
  if (!load(filename))
    exit(1);
}


//  ԡ󥹥ȥ饯   Image(const Image& src)
//      :  ԡImage饹֥
Image::Image(const Image& src)
{
    height = src.height;
    width = src.width;
    xcenter = src.xcenter;
    ycenter = src.ycenter;

#if __GNUC__ < 3
    data = new int[width * height];
#else
    data = new(std::nothrow) int[width * height];
#endif

    if (data == 0) {
        cerr << "cannot allocate memories (Image)." << endl;
        exit(1);
    }
    memcpy(data, src.data, sizeof(int) * width * height);
}

//  Image(const Image& src)
//      :  ԡImage饹֥
//      :  黻
Image& Image::operator =(const Image& src)
{
    // 
    if (this == &src)
        return *this;

    delete [] data;

    height = src.height;
    width = src.width;
    xcenter = src.xcenter;
    ycenter = src.ycenter;

#if __GNUC__ < 3
    data = new int[width * height];
#else
    data = new(std::nothrow) int[width * height];
#endif

    if (data == 0) {
        cerr << "cannot allocate memories (Image)." << endl;
        exit(1);
    }
    memcpy(data, src.data, sizeof(int) * width * height);

    return *this;
}

//  void  set_size(int xs, int ys)
//      :  ν, 
//      :  
void Image::set_size(int xs, int ys)
{
  assert(xs > 0 && ys > 0);
  
  if (data) delete [] data;

  width   = xs;
  height  = ys;
  xcenter = xs / 2;
  ycenter = ys / 2;

#if __GNUC__ < 3
  data = new int [ys * xs];
#else
  data = new(std::nothrow) int [ys * xs];
#endif
  
  if (data == 0) {
    cerr << "cannot allocate memories (Image)." << endl;
    exit(1);
  }
}

// ¿ɸβͤμФ
// double realget(double x, double y)
//      :  ɸ(¿)
//      :  
//      :  ¿ɸβͤμФ
double Image::realget(double x, double y) const
{
  double  s = ycenter - x;
  double  t = xcenter + y;
  double  sn,tn;
  double  g = 0;

  assert((s >= 0) && (s < height-1) && (t >= 0) && (t < width-1));
  
  s = x - floor(x);
  t = y - floor(y);

  sn = 1 - s;
  tn = 1 - t;

  g = get((int)x, (int)y) * s * t + get((int)x+1, (int)y) * sn * t +
      get((int)x, (int)y+1) * s * tn + get((int)x+1, (int)y+1) * sn * tn;

  return g;
  
}

//  int save(const char *filename)
//      :  pgmΥե̾
//      :   1,  0
//      :   pgmǥե¸
int Image::save(const char *filename) const
{
  FILE  *fp;
  int   i;

  if ((fp = fopen(filename, "w")) == NULL)  return 0;

  fprintf(fp, "P5\n");
  fprintf(fp, "%d %d\n", width, height);
  fprintf(fp, "255\n");

  for (i = 0; i < width * height; i++)
    fputc((0x00ff & data[i]), fp);
  fclose(fp);

  return 1;
}


//  int load(const char *filename)
//      :  pgmΥե̾
//      :   1,  0
//      :   pgmΥե뤫ɤ߹
int Image::load(const char *filename)
{
  FILE  *fp;
  int   xs, ys;
  int   maxlevel;
  char  magic[BUFSIZ];
  char  c;
  int   count = 0;

  //  եΥץ
  if ((fp = fopen(filename, "r")) == NULL) {
    cerr << "file not found (" << filename << ")." << endl;
    return 0;
  }
  //  إåʬɤ߹
  if (!head_jump(fp, magic, xs, ys, maxlevel)){
    cerr << "not pgm file." << endl;
    return 0;
  }
   
  // pgmեγǧ
  if (magic[0] != 'P' || magic[1] != '5'){
    cerr << "not pgm file" << endl;
    return 0;
  }
  // 饹Υκ
  if (width != xs || height != ys)
    set_size(xs, ys);

  while (count < width * height){
    c = fgetc(fp);
    data[count] = 0x00ff & c;
    count++;
  }

  // եΥ
  fclose(fp);
  
  return 1;
}


//  void copy_area(Image &Im0, int stu, int stv, 
//                 int width, int height, Image &Im)
//      :  , (饹ɸ), ڤȴ, 
//      :  ʤ
//      :  Im0stu,stvwidthheightʬκ
void copy_area(Image &Im0, int stu, int stv, 
           int width, int height, Image &Im)
{
  int i,j;

  Im.set_size(width, height);
  for (i = 0; i < height; i++)
    for (j = 0; j < width; j++)
      {
        Im.rput(j, i, Im0.rget(stu + j, stv + i));
      }
}


//////////////////
//
//  CImage 饹
//
//////////////////

//  󥹥ȥ饯   CImage(int xs, int ys)
//      :  ν, 
CImage::CImage(int xs, int ys)
{

  assert(xs > 0 && ys > 0);

  width   = xs;
  height  = ys;
  xcenter = xs / 2;
  ycenter = ys / 2;
  
#if __GNUC__ < 3
  data = new Color[width * height];
#else
  data = new(std::nothrow) Color[width * height];
#endif

  if (data == 0) {
    cerr <<"cannot allocate memories (CImage)." <<endl;
    exit(1);
  }
}


//  󥹥ȥ饯   CImage(const char *filename)
//      :  ppmβեΥե̾
CImage::CImage(const char *filename)
{
  width = -1; height = -1; //2005/11/25 ɲá  줬ʤwidth == xs ⤷height == ys Ωľ礬ꡤξ硤dataݤʤ
  data = NULL;
  if (!load(filename))
    exit(1);
}


//  ԡ󥹥ȥ饯   CImage(const CImage& src)
//      :  ԡCImage饹֥
CImage::CImage(const CImage& src)
{
    height = src.height;
    width = src.width;
    xcenter = src.xcenter;
    ycenter = src.ycenter;

#if __GNUC__ < 3
    data = new Color[width * height];
#else
    data = new(std::nothrow) Color[width * height];
#endif

    if (data == 0) {
        cerr << "cannot allocate memories (CImage)." << endl;
        exit(1);
    }
    memcpy(data, src.data, sizeof(Color) * width * height);
}


//  CImage(const CImage& src)
//      :  ԡCImage饹֥
//      :  黻
CImage& CImage::operator =(const CImage& src)
{
    // 
    if (this == &src)
        return *this;

    delete [] data;

    height = src.height;
    width = src.width;
    xcenter = src.xcenter;
    ycenter = src.ycenter;

#if __GNUC__ < 3
    data = new Color[width * height];
#else
    data = new(std::nothrow) Color[width * height];
#endif

    if (data == 0) {
        cerr << "cannot allocate memories (CImage)." << endl;
        exit(1);
    }
    memcpy(data, src.data, sizeof(Color) * width * height);

    return *this;
}


//  void set_size(int xs, int ys)
//      :  ν, 
//      :  
void CImage::set_size(int xs, int ys)
{
  assert(xs > 0 && ys > 0);
  
  if (data != NULL)  delete [] data; //2005/1125 Х data == NULLȤʤäƤ

  width   = xs;
  height  = ys;
  xcenter = xs / 2;
  ycenter = ys / 2;

#if __GNUC__ < 3
  data = new Color[width * height];
#else
  data = new(std::nothrow) Color[width * height];
#endif
  
  if (data == 0) {
    cerr <<"cannot allocate memories (CImage)." <<endl;
    exit(1);
  }
}

// ¿ɸβͤμФ
// void realget(double x, double y, double &r, double &g, double &b)
//
//      :  ɸ(¿), r,g,b (), ߿
//      :  ʤ
//      :  ¿ɸβͤμФ
void CImage::realget(double x, double y, double &r, double &g, double &b)
{
  double  s = ycenter - x;
  double  t = xcenter + y;
  double  sn,tn;
  Color col1, col2, col3, col4;

  assert((s >= 0) && (s < height-1) && (t >= 0) && (t < width-1));
  
  s = x - floor(x);
  t = y - floor(y);

  sn = 1 - s;
  tn = 1 - t;

  col1 = get((int)x  ,(int)y);
  col2 = get((int)x+1,(int)y);
  col3 = get((int)x  ,(int)y+1);
  col4 = get((int)x+1,(int)y+1);

  r = col1.r * s * t  + col2.r * sn * t +
      col3.r * s * tn + col4.r * sn * tn;
  g = col1.g * s * t  + col2.g * sn * t +
      col3.g * s * tn + col4.g * sn * tn;
  b = col1.b * s * t  + col2.b * sn * t +
      col3.b * s * tn + col4.b * sn * tn;

  return;
}

//  int save(const char *filename)
//      :  ppmΥե̾
//      :   1,  0
//      :  ppmΥե¸
int CImage::save(const char *filename) const
{
  FILE  *fp;
  int   i;

  if ((fp = fopen(filename, "w")) == NULL)
    return 0;

  fprintf(fp, "P6\n");
  fprintf(fp, "%d %d\n", width, height);
  fprintf(fp, "255\n");

  for (i = 0; i < width * height; i++){
    fputc((0x00ff & data[i].r), fp);
    fputc((0x00ff & data[i].g), fp);
    fputc((0x00ff & data[i].b), fp);
  }
  fclose(fp);

  return 1;
}


//  int load(const char *filename)
//      :  ppmΥե̾
//      :   1,  0
//      :  ppmΥե뤫ɤ߹
int CImage::load(const char *filename)
{
  FILE  *fp;
  int   xs, ys;
  int   maxlevel;
  char  magic[BUFSIZ];
//    char  c;
  int   count = 0;

  //  եΥץ
  if ((fp = fopen(filename, "r")) == NULL) {
    cerr << "file not found (" << filename << ")." << endl;
    return 0;
  }

  //  إåʬɤ߹
  if (!head_jump(fp, magic, xs, ys, maxlevel)){
    cerr << "not ppm file." << endl;
    return 0;
  }

  // ppmեγǧ
  if (magic[0] != 'P' || magic[1] != '6'){
    cerr << "not ppm file." << endl;
    return 0;
  }
  
  // 饹Υκ
  if (width != xs || height != ys)
    set_size(xs, ys);
  while (count < xs * ys){
    data[count].r = 0x00ff & fgetc(fp);
    data[count].g = 0x00ff & fgetc(fp);
    data[count].b = 0x00ff & fgetc(fp);
    count++;
  }

  // եΥ
  fclose(fp);
  
  return 1;
}



//  void conv_gray(Image &img)
//      :  Image饹β()
//      :  졼Ѵ img ˳Ǽ
void CImage::conv_gray(Image &img)
{
  img.set_size(width, height);
  img.set_center(xcenter, ycenter);
  for (int i=0; i < width * height; i++) 
    img.data[i] = (int) (data[i].r + data[i].g + data[i].b)/3;
} 



//  void copy_area(CImage &Im0, int stu, int stv, 
//                 int width, int height, CImage &Im)
//      :  , (饹ɸ), ڤȴ, 
//      :  ʤ
//      :  Im0stu,stvwidthheightʬκ
void copy_area(CImage &Im0, int stu, int stv, 
           int width, int height, CImage &Im)
{
  int i,j;

  Im.set_size(width, height);
  for (i = 0; i < height; i++)
    for (j = 0; j < width; j++)
      {
        Im.rput(j, i, Im0.rget(stu + j, stv + i));
      }
}
