/************************************************
 Image.h   Ver 1.0

             created  by T.Sakamoto (2000/09/11) 
	     modified by H.Kamo     (2000/10/02)
	     modified by H.Kamo     (2000/10/06)
	     modified by K.Suzukawa (2005/11/22)
*************************************************/

#ifndef _IMAGE_H_
#define _IMAGE_H_

#if __GNUC__ < 3
# include <stdio.h>
# include <iostream.h>
# include <stdlib.h>
#else
# include <iostream>
# include <fstream>
# include <cstdio>
# include <cstdlib>
#endif
#include <assert.h>
#include <ctype.h>
#include "ImageDefaults.h"


//////////////////
//
//  Image 饹
//
//////////////////

class Image
{
 protected:
  int  *data;               /* ХåեΥݥ */
  int  height, width;       /* βȽĤΥ   */
  int  xcenter, ycenter;    /* 濴ɸ         */

 public:
  // 󥹥ȥ饯
  Image(int xs = WIDTH, int ys = HEIGHT);
  Image(const char *filename);

  // ԡ󥹥ȥ饯黻
  Image(const Image& src);
  Image& operator =(const Image& src);

  // ǥȥ饯
  virtual ~Image() { delete [] data; data = 0;};

  // 
  void fill(int g = 0) {
    for (int i = 0; i < width * height; i++)
      data[i] = g;
  };

  // , 濴֤κ
  void set_size(int xs , int ys);
  void set_center(int xc , int yc){
    xcenter = xc;
    ycenter = yc;
  };

  // , 濴֤μФ
  int xl() const { return width; };
  int yl() const { return height; };
  int xc() const { return xcenter; };
  int yc() const { return ycenter; };

  // ͤΥå
  void put(int x,int y,int g) { 
    int  s = ycenter - x;
    int  t = xcenter + y;
    assert((s >= 0) && (s < height) && (t >= 0) && (t < width));
    data[s * width + t] = g;
  };
   
  // ͤμФ
  int get(int x,int y) const {
    int  s = ycenter - x;
    int  t = xcenter + y;
    assert((s >= 0) && (s < height) && (t >= 0) && (t < width));
    return data[s * width + t];
  };

  // ͤΥå (饹ɸ)
  void rput(int u, int v, int value) {
    assert((u >= 0) && (u < width) && (v >= 0) && (v < height));
    data[v * width + u] = value;
  }

  // ͤμФ (饹ɸ)
  int rget(int u, int v) const {
    assert((u >= 0) && (u < width) && (v >= 0) && (v < height));
    return data[v * width + u];
  }

  // ¿ɸβͤμФ(ɸ)
  double realget(double x, double y) const;

  // pgmΥե뤫ɤ߹
  int load(const char *filename);

  // pgmǥե˽
  int save(const char *filename) const;

  // եɥ饹
  friend class CImage;
};

// Im0stu,stvwidthheightʬκ
void copy_area(Image &Im0, int stu, int stv, 
	       int width, int height, Image &Im);


//////////////////
//
//  CImage 饹
//
//////////////////

class CImage
{
 protected:
  Color   *data;               /* ХåեΥݥ */
  int     height, width;       /* βȽĤΥ   */
  int     xcenter, ycenter;    /* 濴ɸ         */


 public:
  // 󥹥ȥ饯
  CImage(int xs = WIDTH, int ys = HEIGHT);
  CImage(const char *filename);

  // ԡ󥹥ȥ饯黻
  CImage(const CImage& src);
  CImage& operator =(const CImage& src);

  // ǥȥ饯
  virtual ~CImage() { delete [] data; data = 0;};

  // 
  void fill(const Color &col = BLACK_C){
    for (int i = 0; i < width * height; i++)
      data[i] = col;
  };

  // , 濴֤κ
  void set_size(int xs, int ys);
  void set_center(int xc, int yc){
    xcenter = xc;
    ycenter = yc;
  };

  // , 濴֤μФ
  int xl() const { return width; };
  int yl() const { return height; };
  int xc() const { return xcenter; };
  int yc() const { return ycenter; };

  // ͤ򥻥å
  void put(int x, int y, int r, int g, int b){
    int  s = ycenter - x;
    int  t = xcenter + y;
    assert((s >= 0) && (s < height) && (t >= 0) && (t < width));
    int  i = s * width + t;
    data[i].r = r;
    data[i].g = g;
    data[i].b = b;
  };

  void put(int x, int y, Color c){
    int  s = ycenter - x;
    int  t = xcenter + y;
    assert((s >= 0) && (s < height) && (t >= 0) && (t < width));
    //    int  i = s * width + t;
    //    data[i].r = c.r;
    //    data[i].g = c.g;
    //    data[i].b = c.b;
    data[s * width + t] = c;
  };

  // ͤμФ
  void get(int x, int y, int &r, int &g, int &b) const {
    int  s = ycenter - x;
    int  t = xcenter + y;
    assert((s >= 0) && (s < height) && (t >= 0) && (t < width));
    int  i = s * width + t;
    r = data[i].r;
    g = data[i].g;
    b = data[i].b;
  };

  Color get(int x, int y) const {
    int  s = ycenter - x;
    int  t = xcenter + y;
    assert((s >= 0) && (s < height) && (t >= 0) && (t < width));
    return data[s * width + t];
  };

  // ͤΥå (饹ɸ)
  void rput(int u, int v, int r, int g, int b) {
    assert((u >= 0) && (u < width) && (v >= 0) && (v < height));
    int  i = v * width + u;
    data[i].r = r;
    data[i].g = g;
    data[i].b = b;
  }

  void rput(int u, int v, Color col) {
    assert((u >= 0) && (u < width) && (v >= 0) && (v < height));
    //    int  i = v * width + u;
    //    data[i].r = r;
    //    data[i].g = g;
    //    data[i].b = b;
    data[v * width + u] = col;
  }

  // ͤμФ (饹ɸ)
  void rget(int u, int v, int &r, int &g, int &b) const {
    assert((u >= 0) && (u < width) && (v >= 0) && (v < height));
    int  i = v * width + u;
    r = data[i].r;
    g = data[i].g;
    b = data[i].b;
  }

  Color rget(int u, int v) const {
    assert((u >= 0) && (u < width) && (v >= 0) && (v < height));
    return data[v * width + u];
  }
  
  // ¿ɸβͤμФ(ɸ)
  void realget(double x, double y, double &r, double &g, double &b);

  // ppmΥե뤫ɤ߹
  int load(const char *filename);

  // ppmΥե˽
  int save(const char *filename) const;

  // 졼(Image饹)Ѵ
  void conv_gray(Image &img);
};

// Im0stu,stvwidthheightʬκ
void copy_area(CImage &Im0, int stu, int stv, 
	       int width, int height, CImage &Im);

#endif //_IMAGE_H_
