/************************************************
 JpegUtils.cc   Ver 1.0

             created  by Y. Kanazawa (2002/08/28)
************************************************/

#if __GNUC__ < 3
# include <stdio.h>
#else
# include <iostream>
# include <new>
using std::ostream;
using std::ofstream;
using std::cerr;
using std::endl;
#endif
#include <jpeglib.h>
#include "Image.h"
#include "JpegUtils.h"


static unsigned char buf[4];
static unsigned long power85[5] = { 1L, 85L, 85L*85, 85L*85*85, 85L*85*85*85};
static int outbytes;		/* Number of characters in an output line */

/* read 0-4 Bytes. result: number of bytes read */
static int 
readsomebytes(FILE *in)
{
   register int count, i;

   for (count = 0; count < 4; count++)
   {
      if ((i = getc(in)) == EOF)
	 break;
      else
	 buf[count] = (unsigned char) i;
   }
   return count;
}

static void 
outabyte(unsigned char c, ostream &out)
{    /* output one byte */

   out << c;
   if (out.eof())
   {
      cerr << "output_jpegascii: write error - exit!" << endl;
      exit(1);
   }
   
   if (++outbytes > 63 ||(outbytes == 1 && c == '%') )
   {
      out << endl;
      outbytes = 0;
   }
}

static int
output_ascii85(FILE *in, ostream &out)
{
   register int i, count;
   unsigned long word, v;

   outbytes = 0;
  
   /* 4 bytes read ==> output 5 bytes */
   while ((count = readsomebytes(in)) == 4)
   {
      word = ((unsigned long)(((unsigned int)buf[0] << 8) + buf[1]) << 16) +
	 (((unsigned int)buf[2] << 8) + buf[3]);
      if (word == 0)
	 outabyte('z', out);       /* shortcut for 0 */
      else
	 /* calculate 5 ASCII85 bytes and output them */
	 for (i = 4; i >= 0; i--)
	 {
	    v = word / power85[i];
	    outabyte((unsigned char) (v + '!'), out);
	    word -= v * power85[i];
	 }
   }

   word = 0;

   if (count != 0)
   {   /* 1-3 bytes left */
      for (i = count-1; i >= 0; i--)   /* accumulate bytes */
	 word += (unsigned long)buf[i] << 8 * (3-i);
    
      /* encoding as above, but output only count+1 bytes */
      for (i = 4; i >= 4-count; i--)
      {
	 v = word / power85[i];
	 outabyte((unsigned char) (v + '!'), out);
	 word -= v * power85[i];
      }
   }
   
   out << "~>" << endl;	/* EOD marker */
   return 0;
}


//  int output_mono_jpegascii(Image &im, ofstream &out, int quality)
//      :  Image
//    	       ϥȥ꡼
//             ʼ
//      :   1,  0
//      :  PostscriptǽϤ
int
output_mono_jpegascii(const Image &im, ofstream &out, int quality)
{
   struct jpeg_compress_struct	cinfo;
   struct jpeg_error_mgr	jerr;
   FILE				*tmpf;	/* target file */
   JSAMPROW			row_pointer[1];	/* pointer to JSAMPLE row[s] */
   int				row_stride;	/* physical row width */
   int				i, u, v;
   int				width, height;

   JSAMPLE			*image_buffer;	/* image */

   width = im.xl();
   height = im.yl();
   /* allocate buffer */
#if __GNUC__ < 3
   image_buffer = new JSAMPLE [width * height];
#else
   image_buffer = new(std::nothrow) JSAMPLE [width * height];
#endif
   if (image_buffer == 0)
   {
      cerr << "output_mono_jpegascii(): cannot allocate buffer!" << endl;
      return 0;
   }

   /* set up error handler */
   cinfo.err = jpeg_std_error(&jerr);
   /* initialize the JPEG compression object. */
   jpeg_create_compress(&cinfo);

   /* open temporary file as a stream */
//   if ((tmpf = fopen("temporary.jpg", "wb")) == NULL)
   if ((tmpf = tmpfile()) == NULL)
   {
      cerr << "output_mono_jpegascii(): can't open temporary file!";
      delete [] image_buffer;
      return 0;
   }
   jpeg_stdio_dest(&cinfo, tmpf);

   /* set image size etc. */
   cinfo.image_width = width; 	/* width and height */
   cinfo.image_height = height;
   cinfo.input_components = 1;	/* # of color components per pixel */
   cinfo.in_color_space = JCS_GRAYSCALE; 	/* colorspace of input image */

   jpeg_set_defaults(&cinfo);
   /* set quality */
   jpeg_set_quality(&cinfo, quality, TRUE);

   /* start compressor */
   jpeg_start_compress(&cinfo, TRUE);

   row_stride = width; /* JSAMPLEs per row in image_buffer */

   i = v = 0;
   while (v < height && cinfo.next_scanline < cinfo.image_height)
   {
      for (u = 0; u < width; u++)
	 image_buffer[i++] = im.rget(u, v);

      row_pointer[0] = & image_buffer[cinfo.next_scanline * row_stride];
      (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
      v++;
   }

   /* finish compression */
   jpeg_finish_compress(&cinfo);

   /* rewind file */
   fseek(tmpf, 0L, SEEK_SET);

   /* output jpeg ascii85 data */
   output_ascii85(tmpf, out);
   fclose(tmpf);

   /* release memory. */
   jpeg_destroy_compress(&cinfo);

   delete [] image_buffer;

   return 1;
}


//  int output_jpegascii(CImage &im, ofstream &out, int quality)
//      :  CImage
//    	       ϥȥ꡼
//             ʼ
//      :   1,  0
//      :  PostscriptǽϤ
int
output_jpegascii(const CImage &im, ofstream &out, int quality)
{
   struct jpeg_compress_struct	cinfo;
   struct jpeg_error_mgr	jerr;
   FILE				*tmpf;	/* target file */
   JSAMPROW			row_pointer[1];	/* pointer to JSAMPLE row[s] */
   int				row_stride;	/* physical row width */
   int				i, u, v;
   int				r, g, b, width, height;

   JSAMPLE			*image_buffer;	/* image */

   width = im.xl();
   height = im.yl();
   /* allocate buffer */
#if __GNUC__ < 3
   image_buffer = new JSAMPLE [3 * width * height];
#else
   image_buffer = new(std::nothrow) JSAMPLE [3 * width * height];
#endif
   if (image_buffer == 0)
   {
      cerr << "output_jpegascii(): cannot allocate buffer!" << endl;
      return 0;
   }

   /* set up error handler */
   cinfo.err = jpeg_std_error(&jerr);
   /* initialize the JPEG compression object. */
   jpeg_create_compress(&cinfo);

   /* open temporary file as a stream */
//   if ((tmpf = fopen("temporary.jpg", "wb")) == NULL)
   if ((tmpf = tmpfile()) == NULL)
   {
      cerr << "output_jpegascii(): can't open temporary file!";
      delete [] image_buffer;
      return 0;
   }
   jpeg_stdio_dest(&cinfo, tmpf);

   /* set image size etc. */
   cinfo.image_width = width; 	/* width and height */
   cinfo.image_height = height;
   cinfo.input_components = 3;	/* # of color components per pixel */
   cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */

   jpeg_set_defaults(&cinfo);
   /* set quality */
   jpeg_set_quality(&cinfo, quality, TRUE);

   /* start compressor */
   jpeg_start_compress(&cinfo, TRUE);

   row_stride = width * 3; /* JSAMPLEs per row in image_buffer */

   i = v = 0;
   while (v < height && cinfo.next_scanline < cinfo.image_height)
   {
      for (u = 0; u < width; u++)
      {
	 im.rget(u, v, r, g, b);
	 image_buffer[i++] = r;
	 image_buffer[i++] = g;
	 image_buffer[i++] = b;
      }
      row_pointer[0] = & image_buffer[cinfo.next_scanline * row_stride];
      (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
      v++;
   }

   /* finish compression */
   jpeg_finish_compress(&cinfo);

   /* rewind file */
   fseek(tmpf, 0L, SEEK_SET);

   /* output jpeg ascii85 data */
   output_ascii85(tmpf, out);
   fclose(tmpf);

   /* release memory. */
   jpeg_destroy_compress(&cinfo);

   delete [] image_buffer;

   return 1;
}
